import unittest

from src.pytrivialsql import sql


class TestWhereToString(unittest.TestCase):
    def test_where_dict(self):
        self.assertEqual(
            ("a=?", (1,)), sql._where_dict_to_string({"a": 1}, placeholder="?")
        )
        self.assertEqual(
            ("a=? AND b=?", (1, 2)),
            sql._where_dict_to_string({"a": 1, "b": 2}, placeholder="?"),
        )
        self.assertEqual(
            ("a=? AND b IN ('A', 'B')", (1,)),
            sql._where_dict_to_string({"a": 1, "b": {"A", "B"}}, placeholder="?"),
        )
        self.assertEqual(
            ("a IS NULL", ()), sql._where_dict_to_string({"a": None}, placeholder="?")
        )

    def test_where_arr(self):
        self.assertEqual(
            ("(a=?)", (1,)), sql._where_arr_to_string([{"a": 1}], placeholder="?")
        )
        self.assertEqual(
            ("(a=?) OR (b=?)", (1, 2)),
            sql._where_arr_to_string([{"a": 1}, {"b": 2}], placeholder="?"),
        )
        self.assertEqual(
            ("(a=? AND c=?) OR (b=?)", (1, 3, 2)),
            sql._where_arr_to_string([{"a": 1, "c": 3}, {"b": 2}], placeholder="?"),
        )

    def test_where_tuple(self):
        self.assertEqual(
            ("a like ?", (1,)),
            sql._where_tup_to_string(("a", "like", 1), placeholder="?"),
        )
        self.assertEqual(
            ("NOT (a IS NULL)", ()),
            sql._where_tup_to_string(("NOT", {"a": None}), placeholder="?"),
        )
        self.assertIsNone(sql._where_tup_to_string(("a", "like"), placeholder="?"))
        self.assertIsNone(
            sql._where_tup_to_string(("a", "like", "b", "c"), placeholder="?")
        )

    def test_where(self):
        self.assertEqual(
            ("(a=?) OR (a like ?)", ("blah", "something else")),
            sql._where_to_string(
                [{"a": "blah"}, ("a", "like", "something else")], placeholder="?"
            ),
        )
        self.assertEqual(
            ("(a=? AND b=?) OR (a like ?)", ("blah", "bleeh", "something else")),
            sql._where_to_string(
                [{"a": "blah", "b": "bleeh"}, ("a", "like", "something else")],
                placeholder="?",
            ),
        )

    def test_postgres_placeholder(self):
        self.assertEqual(
            ("(a=%s) OR (a like %s)", ("blah", "something else")),
            sql._where_to_string(
                [{"a": "blah"}, ("a", "like", "something else")], placeholder="%s"
            ),
        )
        self.assertEqual(
            ("(a=%s AND b=%s) OR (a like %s)", ("blah", "bleeh", "something else")),
            sql._where_to_string(
                [{"a": "blah", "b": "bleeh"}, ("a", "like", "something else")],
                placeholder="%s",
            ),
        )


class TestCreate_q(unittest.TestCase):
    def test_string_rep(self):
        self.assertEqual(
            sql.create_q(
                "table_name",
                [
                    "column ID PRIMARY KEY",
                    "prop TEXT",
                    "propb INTEGER",
                    "created DATETIME DEFAULT CURRENT_TIMESTAMP NOT NULL",
                ],
            ),
            "CREATE TABLE IF NOT EXISTS table_name(column ID PRIMARY KEY, prop TEXT, propb INTEGER, created DATETIME DEFAULT CURRENT_TIMESTAMP NOT NULL)",
        )


class TestInsert_q(unittest.TestCase):
    def test_string_rep(self):
        self.assertEqual(
            sql.insert_q("table_name", prop="Blah!"),
            ("INSERT INTO table_name (prop) VALUES (?)", ("Blah!",)),
        )
        self.assertEqual(
            sql.insert_q("table_name", prop="Blah!", propb=12),
            ("INSERT INTO table_name (prop, propb) VALUES (?, ?)", ("Blah!", 12)),
        )

    def test_returning_clause(self):
        self.assertEqual(
            sql.insert_q("users", prop="foo", propb="bar", RETURNING=["id"]),
            (
                "INSERT INTO users (prop, propb) VALUES (?, ?) RETURNING id",
                ("foo", "bar"),
            ),
        )
        self.assertEqual(
            sql.insert_q("users", prop="foo", propb="bar", RETURNING=["id", "created"]),
            (
                "INSERT INTO users (prop, propb) VALUES (?, ?) RETURNING id, created",
                ("foo", "bar"),
            ),
        )
        self.assertEqual(
            sql.insert_q("users", prop="foo", propb="bar", RETURNING="*"),
            (
                "INSERT INTO users (prop, propb) VALUES (?, ?) RETURNING *",
                ("foo", "bar"),
            ),
        )

    def test_different_placeholder(self):
        self.assertEqual(
            sql.insert_q("table_name", prop="Blah!", propb=12, placeholder="%s"),
            ("INSERT INTO table_name (prop, propb) VALUES (%s, %s)", ("Blah!", 12)),
        )

    def test_placeholder_doesnt_mutate_external(self):
        props = {"prop": "Blah!", "propb": 12, "placeholder": "%s"}
        self.assertEqual(
            sql.insert_q("table_name", **props),
            ("INSERT INTO table_name (prop, propb) VALUES (%s, %s)", ("Blah!", 12)),
        )


class TestDelete_q(unittest.TestCase):
    def test_string_rep(self):
        self.assertEqual(
            ("DELETE FROM table_name WHERE id=?", (1,)),
            sql.delete_q("table_name", where={"id": 1}),
        )

    def test_placeholder_change(self):
        self.assertEqual(
            ("DELETE FROM table_name WHERE id=%s", (1,)),
            sql.delete_q("table_name", where={"id": 1}, placeholder="%s"),
        )


class TestUpdate_q(unittest.TestCase):
    def test_string_rep(self):
        self.assertEqual(
            ("UPDATE table_name SET prop=?", ("Bleeh!",)),
            sql.update_q("table_name", prop="Bleeh!"),
        )
        self.assertEqual(
            ("UPDATE table_name SET prop=? WHERE id=?", ("Bleeh!", 1)),
            sql.update_q("table_name", prop="Bleeh!", where={"id": 1}),
        )

    def test_different_placeholder(self):
        self.assertEqual(
            (
                "UPDATE table_name SET prop=%s,proq=%s WHERE id=%s",
                ("Bleeh!", "blah", 1),
            ),
            sql.update_q(
                "table_name",
                prop="Bleeh!",
                proq="blah",
                where={"id": 1},
                placeholder="%s",
            ),
        )


class TestSelect_q(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(
            sql.select_q("table_name", ["id", "prop", "propb"]),
            ("SELECT id, prop, propb FROM table_name", ()),
        )

    def test_where(self):
        self.assertEqual(
            sql.select_q("table_name", ["id", "prop", "propb"], where={"a": 1}),
            ("SELECT id, prop, propb FROM table_name WHERE a=?", (1,)),
        )

    def test_where_with_placeholder_change(self):
        self.assertEqual(
            sql.select_q(
                "table_name", ["id", "prop", "propb"], where={"a": 1}, placeholder="%s"
            ),
            ("SELECT id, prop, propb FROM table_name WHERE a=%s", (1,)),
        )

    def test_order_by(self):
        self.assertEqual(
            sql.select_q(
                "table_name", ["id", "prop", "propb"], where={"a": 1}, order_by="prop"
            ),
            ("SELECT id, prop, propb FROM table_name WHERE a=? ORDER BY prop", (1,)),
        )

    def test_limit(self):
        self.assertEqual(
            sql.select_q(
                "table_name", ["id", "prop", "propb"], where={"a": 1}, limit=2
            ),
            ("SELECT id, prop, propb FROM table_name WHERE a=? LIMIT 2", (1,)),
        )

    def test_offset(self):
        self.assertEqual(
            sql.select_q(
                "table_name", ["id", "prop", "propb"], where={"a": 1}, offset=5
            ),
            ("SELECT id, prop, propb FROM table_name WHERE a=? OFFSET 5", (1,)),
        )

    def test_distinct(self):
        self.assertEqual(
            sql.select_q(
                "table_name", ["id", "prop", "propb"], where={"a": 1}, distinct="prop"
            ),
            ("SELECT DISTINCT (prop) id, prop, propb FROM table_name WHERE a=?", (1,)),
        )
        self.assertEqual(
            sql.select_q(
                "table_name",
                ["id", "prop", "propb"],
                where={"a": 1},
                distinct_on="prop",
            ),
            (
                "SELECT DISTINCT ON (prop) id, prop, propb FROM table_name WHERE a=?",
                (1,),
            ),
        )
        self.assertEqual(
            sql.select_q(
                "table_name",
                ["id", "prop", "propb"],
                where={"a": 1},
                distinct_on=["prop", "propb"],
            ),
            (
                "SELECT DISTINCT ON (prop, propb) id, prop, propb FROM table_name WHERE a=?",
                (1,),
            ),
        )



import json

import psycopg

from . import sql

_JSON_TYPES = {list, dict}


class Postgres:
    def __init__(self, db_url, autocommit=False):
        self._autocommit = autocommit
        self._url = db_url
        self._connect()

    def _commit(self):
        if not self._autocommit:
            self._conn.commit()

    def _connect(self):
        self._conn = psycopg.connect(self._url, autocommit=self._autocommit)

    def _reconnect(self):
        self.close()
        self._connect()

    def close(self):
        self._conn.close()

    def exec(self, query, args=None):
        try:
            with self._conn.cursor() as cur:
                cur.execute(query, args)
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def execs(self, query_args_pairs):
        try:
            with self._conn.cursor() as cur:
                for q, qargs in query_args_pairs:
                    cur.execute(q, qargs)
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def drop(self, *table_names):
        try:
            with self._conn.cursor() as cur:
                for tbl in table_names:
                    cur.execute(sql.drop_q(tbl))
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def create(self, table_name, props):
        with self._conn.cursor() as cur:
            cur.execute(sql.create_q(table_name, props))
            self._commit()
            return True

    def add_column(self, table_name, col):
        with self._conn.cursor() as cur:
            cur.execute(sql.add_column_q(table_name, col))
            self._commit()
            return True

    def select(
        self,
        table_name,
        columns,
        distinct=None,
        where=None,
        order_by=None,
        limit=None,
        offset=None,
        transform=None,
    ):
        try:
            with self._conn.cursor() as cur:
                if columns is None:
                    columns = "*"
                if type(columns) is str:
                    columns = [columns]
                query, args = sql.select_q(
                    table_name,
                    columns,
                    where=where,
                    distinct_on=distinct,
                    order_by=order_by,
                    limit=limit,
                    offset=offset,
                    placeholder="%s",
                )
                cur.execute(query, args)
                cols = [col.name for col in cur.description]
                res = (dict(zip(cols, vals)) for vals in cur.fetchall())
                if transform is not None:
                    return [transform(el) for el in res]
                return list(res)
        except Exception as e:
            self._reconnect()
            raise e

    def insert(self, table_name, **args):
        global _JSON_TYPES
        returning = args.get("returning", args.get("RETURNING", None))
        try:
            with self._conn.cursor() as cur:
                for k, v in args.items():
                    if k in {"returning", "RETURNING"}:
                        if type(v) is str:
                            args[k] = [v]
                    elif type(v) in _JSON_TYPES:
                        args[k] = json.dumps(v)
                args["placeholder"] = "%s"
                q, qargs = sql.insert_q(table_name, **args)
                cur.execute(q, qargs)
                if returning is None:
                    res = None
                else:
                    res = cur.fetchone()
                    if len(res) == 1:
                        res = res[0]
                    else:
                        cols = [col.name for col in cur.description]
                        res = dict(zip(cols, res))
            self._commit()
            return res
        except Exception as e:
            self._reconnect()
            raise e

    def update(self, table_name, bindings, where):
        global _JSON_TYPES
        binds = {"placeholder": "%s", **bindings}
        try:
            with self._conn.cursor() as cur:
                for k, v in binds.items():
                    if type(v) in _JSON_TYPES:
                        binds[k] = json.dumps(v)
                q, args = sql.update_q(table_name, where=where, **binds)
                cur.execute(q, args)
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def delete(self, table_name, where):
        try:
            with self._conn.cursor() as cur:
                cur.execute(*sql.delete_q(table_name, where=where, placeholder="%s"))
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e



import tempfile
import unittest

from src.pytrivialsql import sqlite


class TestDBInteraction(unittest.TestCase):
    def test_basic_interactions(self):
        with tempfile.NamedTemporaryFile(suffix=".db") as f:
            db = sqlite.Sqlite3(f.name)
            self.assertTrue(db.create("a_table", [
                "id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL",
                "a_column TEXT", "a_number_column INTEGER",
                "a_boolean_column INTEGER",
                "created DATETIME DEFAULT CURRENT_TIMESTAMP NOT NULL"
            ]))
            rwid = db.insert("a_table", a_column="A Value", a_number_column=42, a_boolean_column=True)
            rw = db.select("a_table", "*", where={"id": rwid})



import os
import unittest

import psycopg
from src.pytrivialsql import postgres


class TestDBInteraction(unittest.TestCase):
    def test_basic_interactions(self):
        DB = postgres.Postgres(os.environ["POSTGRES_URL"])
        DB.create(
            "a_table",
            [
                "id BIGSERIAL PRIMARY KEY NOT NULL",
                "a_column TEXT",
                "another_column BOOLEAN",
                "a_number_column INTEGER",
                "a_json_column JSONB DEFAULT '[]'::jsonb",
                "another_json_column JSONB DEFAULT '{}'::jsonb",
                "created TIMESTAMP WITHOUT TIME ZONE DEFAULT NOW()",
            ],
        )
        self.assertEqual([], DB.select("a_table", "*"))
        res = DB.insert(
            "a_table",
            a_column="Blah blah",
            another_column=True,
            a_number_column=42,
            a_json_column=[1, 2, 3],
            another_json_column={"a": 1, "b": 2, "c": 3},
            RETURNING="*",
        )
        self.assertEqual(res["a_column"], "Blah blah")
        self.assertEqual(res["a_number_column"], 42)
        self.assertEqual(res["a_json_column"], [1, 2, 3])
        self.assertEqual(res["another_json_column"], {"a": 1, "b": 2, "c": 3})
        self.assertEqual(
            [{"a_json_column": [1, 2, 3]}],
            DB.select("a_table", "a_json_column", where={"id": res["id"]}),
        )
        DB.update("a_table", {"a_json_column": [3, 2, 1]}, where={"id": res["id"]})
        self.assertEqual(
            [{"a_json_column": [3, 2, 1]}],
            DB.select("a_table", "a_json_column", where={"id": res["id"]}),
        )
        DB.update(
            "a_table",
            {"another_column": False, "a_column": "a row"},
            where={"id": res["id"]},
        )
        select_res = DB.select(
            "a_table", ["another_column", "a_column"], where={"id": res["id"]}
        )
        self.assertEqual(select_res[0]["a_column"], "a row")
        self.assertFalse(select_res[0]["another_column"])

        rid = DB.insert("a_table", a_column="another row", RETURNING="id")
        self.assertIsInstance(rid, int)

        rlist = DB.insert(
            "a_table",
            a_column="another row",
            RETURNING=["id", "a_column", "a_json_column"],
        )
        self.assertIsInstance(rlist, dict)

        DB.drop("a_table")
        with self.assertRaises(psycopg.errors.UndefinedTable):
            DB.select("a_table", "*")
        DB.close()



import sqlite3

from . import sql


class Sqlite3:
    def __init__(self, db_path):
        self.path = db_path
        self._conn = sqlite3.connect(
            self.path, check_same_thread=not self.is_threadsafe()
        )

    def exec(self, query, args=None):
        with self._conn as cur:
            cur.execute(query, args)

    def execs(self, query_args_pairs):
        with self._conn as cur:
            for q, qargs in query_args_pairs:
                cur.execute(q, qargs)

    def is_threadsafe(self):
        mem = sqlite3.connect("file::memory:?cache=shared")
        cur = mem.execute(
            "select * from pragma_compile_options where compile_options like 'THREADSAFE=%'"
        )
        res = cur.fetchall()
        cur.close()
        try:
            return res[0][0].split("=")[1] == "1"
        except Exception:
            False

    def drop(self, *table_names):
        with self._conn as cur:
            for tbl in table_names:
                cur.execute(sql.drop_q(tbl))

    def create(self, table_name, props):
        try:
            with self._conn as cur:
                cur.execute(sql.create_q(table_name, props))
                return True
        except Exception:
            return False

    def add_column(self, table_name, col):
        try:
            with self._conn as cur:
                cur.execute(sql.add_column_q(table_name, col))
                return True
        except Exception:
            return False

    def select(
        self,
        table_name,
        columns,
        distinct=None,
        where=None,
        order_by=None,
        limit=None,
        offset=None,
        transform=None,
    ):
        with self._conn as cur:
            c = cur.cursor()
            if columns is None or columns == "*":
                columns = [
                    el[1]
                    for el in c.execute(f"PRAGMA table_info({table_name})").fetchall()
                ]
            if not columns:
                raise Exception(f"No such table {table_name}")
            elif isinstance(columns, str):
                columns = [columns]
            query, args = sql.select_q(
                table_name,
                columns,
                where=where,
                distinct=distinct,
                order_by=order_by,
                limit=limit,
                offset=offset,
            )
            c.execute(query, args)
            res = (dict(zip(columns, vals)) for vals in c.fetchall())
            if transform is not None:
                return [transform(el) for el in res]
            return list(res)

    def insert(self, table_name, **args):
        with self._conn as cur:
            c = cur.cursor()
            c.execute(*sql.insert_q(table_name, **args))
            return c.lastrowid

    def update(self, table_name, bindings, where):
        with self._conn as cur:
            c = cur.cursor()
            q, args = sql.update_q(table_name, where=where, **bindings)
            c.execute(q, args)

    def delete(self, table_name, where):
        with self._conn as cur:
            c = cur.cursor()
            c.execute(*sql.delete_q(table_name, where=where))
